/*
 * jEdit editor settings:
 * :mode=java:tabSize=4:indentSize=4:noTabs=true:maxLineLen=0:
 *
 * JIndexHoldable.java
 * Copyright (C) 1999 2000 Dirk Moebius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jindex;

import java.awt.*;
import java.awt.event.*;
import java.net.*;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;

import org.gjt.sp.jedit.*;
import org.gjt.sp.jedit.msg.ShowTokenInfo;
import org.gjt.sp.jedit.msg.ViewURL;
import org.gjt.sp.jedit.gui.DockableWindow;
import org.gjt.sp.util.Log;

import JIndexPlugin;


/**
 * This is the component, that displays the alphabetically sorted list of
 * index keywords, the search field and the "found entries" list.
 */
public class JIndexHoldable extends JPanel implements EBComponent, DockableWindow {

    private static final Dimension KEY_LIST_SIZE = new Dimension(300,400);
    private static final Dimension TOPIC_LIST_SIZE = new Dimension(300,150);
    private static final Insets vertSpaceTop = new Insets(5,5,0,5);
    private static final Insets vertSpaceNone = new Insets(0,5,0,5);
    private static final Insets vertSpaceBoth = new Insets(5,5,5,5);
    private static final int fixedCellHeight = new JLabel("Mg_").getPreferredSize().height;

    final static ImageIcon classIcon = new ImageIcon(JIndexHoldable.class.getResource("images/class.gif"));
    final static ImageIcon interIcon = new ImageIcon(JIndexHoldable.class.getResource("images/interf.gif"));
    final static ImageIcon constrIcon = new ImageIcon(JIndexHoldable.class.getResource("images/constr.gif"));
    final static ImageIcon methodIcon = new ImageIcon(JIndexHoldable.class.getResource("images/method.gif"));
    final static ImageIcon fieldIcon = new ImageIcon(JIndexHoldable.class.getResource("images/field.gif"));

    private JIndex index;
    private JTextField searchField;
    private HelpfulJList keyList;
    private HelpfulJList topicList;
    private TopicListModel tlmodel;
    private JLabel status;
    private JSplitPane splitpane;
    private int keyselpos = -1;
    private int topicselpos = -1;


    public JIndexHoldable() {
        // ************ create components ************

        // search field
        searchField = new JTextField();
        searchField.getDocument().addDocumentListener(new DocumentListener() {
            public void insertUpdate(DocumentEvent e) { search(); }
            public void removeUpdate(DocumentEvent e) { search(); }
            public void changedUpdate(DocumentEvent e) { search(); }
            private void search() {
                if (index == null) {
                    setStatusText(prop("status.notloaded"));
                    return;
                }
                String searchstring = searchField.getText().trim();
                setSearchResult(index.search(searchstring, false));
            }
        });

        // label for search field
        JLabel l1 = new JLabel(prop("search.label"));
        l1.setDisplayedMnemonic(prop("search.mnemonic").charAt(0));
        l1.setLabelFor(searchField);

        // list of keywords
        keyList = new HelpfulJList(new KeyListModel());
        keyList.setPrototypeCellValue("W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_W_");
        keyList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        keyList.addListSelectionListener(new ListSelectionListener() {
            public void valueChanged(ListSelectionEvent e) {
                if (index == null) return;
                keyselpos = keyList.getSelectedIndex();
                if (keyselpos < 0 || keyselpos >= index.getNumKeywords()) return;
                updateTopicList(index.getEntriesAt(keyselpos));
            }
        });

        keyList.addMouseListener(new MouseAdapter() {
            public void mouseClicked(MouseEvent e) {
                if (e.getClickCount()==2)
                    if (topicList.getModel().getSize() == 1)
                        showDocOnSelectedEntry();
                    else
                        topicList.requestFocus();
            }
        });

        // label for list of keywords
        JLabel l2 = new JLabel(prop("index.label"));
        l2.setDisplayedMnemonic(prop("index.mnemonic").charAt(0));
        l2.setLabelFor(keyList);

        // scrollpane for list of keywords
        JScrollPane scrKeyList = new JScrollPane(keyList);
        scrKeyList.setPreferredSize(KEY_LIST_SIZE);
        scrKeyList.setColumnHeaderView(l2);

        // list of topics found
        tlmodel = new TopicListModel();

        topicList = new HelpfulJList(tlmodel);
        topicList.setFixedCellHeight(fixedCellHeight);
        topicList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        topicList.setCellRenderer(new TopicListCellRenderer());
        topicList.addMouseListener(new MouseAdapter() {
            public void mouseClicked(MouseEvent e) {
                if (e.getClickCount()==2) {
                    topicselpos = topicList.locationToIndex(e.getPoint());
                    showDocOnSelectedEntry();
                }
            }
        });

        // label for list of topics
        JLabel l3 = new JLabel(prop("topics.label"));
        l3.setDisplayedMnemonic(prop("topics.mnemonic").charAt(0));
        l3.setLabelFor(topicList);

        // scrollpane for list of topics
        JScrollPane scrTopicList = new JScrollPane(topicList);
        scrTopicList.setPreferredSize(TOPIC_LIST_SIZE);
        scrTopicList.setColumnHeaderView(l3);

        // splitpane with keylist and topiclist
        splitpane = new JSplitPane(JSplitPane.VERTICAL_SPLIT, true, scrKeyList, scrTopicList);
        splitpane.setOneTouchExpandable(true);

        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                String divLoc = jEdit.getProperty("jindex.dividerLocation");
                if (divLoc != null)
                    splitpane.setDividerLocation(Integer.parseInt(divLoc));
            }
        });

        // status label
        status = new JLabel(prop("status.label"));
        status.setBorder(new BevelBorder(BevelBorder.LOWERED));
        status.setFont(new Font("Dialog", Font.PLAIN, 10));

        // ************ general layout ************

        this.setLayout(new BorderLayout());

        // add search field with label
        Box top = Box.createHorizontalBox();
        top.add(l1);
        top.add(searchField);
        this.add(top, BorderLayout.NORTH);

        // add splitpane with keyword list and topic list
        this.add(splitpane, BorderLayout.CENTER);

        // add status label
        this.add(status, BorderLayout.SOUTH);

        // ************ misc setup ************

        // set default size
        setSize(new Dimension(220,500));  // faster than pack()

        // we like to be notified about JIndex and status changes:
        EditBus.addToBus(this);

        // load index:
        JIndexLoader jil = new JIndexLoader();
        jil.start();
        index = jil.getIndex(); // note: the index may still be null

        // initialize both lists:
        init();
    }


    public void handleMessage(EBMessage message) {
        if (message instanceof StatusChange) {
            int newstatus = ((StatusChange)message).getStatus();
            if (newstatus == JIndexLoader.STATUS_LOADING) {
                setBusy(true);
                setStatusText(prop("status.loading"));
            } else {
                setBusy(false);
                if (newstatus != JIndexLoader.STATUS_NOT_EXISTS &&
                    newstatus != JIndexLoader.STATUS_OK) {
                    // there was some error
                    setStatusText(prop("status.error"));
                }
            }
        }
        else if (message instanceof JIndexChange) {
            index = ((JIndexChange)message).getJIndex();
            Log.log(Log.DEBUG, this, "got index change");
            init();
            setStatusText(prop("status.ready"));
        }
    }


    public String getName() {
        return JIndexPlugin.NAME;
    }


    public Component getComponent() {
        return this;
    }


    /**
     * Invoked when the component is removed; saves the position of the
     * split pane divider
     */
    public void removeNotify() {
        super.removeNotify();
        jEdit.setProperty("jindex.dividerLocation", Integer.toString(splitpane.getDividerLocation()));
    }


    private void init() {
        keyList.setModel(new KeyListModel());
        keyselpos = -1;
        topicselpos = -1;
        tlmodel.setEntries(null);
    }


    private void setStatusText(final String text) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                status.setText(text);
            }
        });
    }


    private void setBusy(boolean busy) {
        if (busy)
            setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
        else
            setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
    }


    private String prop(String key) {
        return jEdit.getProperty("jindex.frame." + key);
    }


    public void search(String searchstring) {
        if (index == null)
            return; // there is no index, probable because of error
        if (searchstring == null)
            return; // nothing to search for

        // search in index for searchstring
        JIndex.SearchResult res = index.search(searchstring, true);
        if (res.entries == null) {
            // nothing found
            setStatusText(jEdit.getProperty(
                "jindex.frame.status.keywordnotfound",
                new Object[] { res.searchstring } ));
            Toolkit.getDefaultToolkit().beep();
        } else {
            setSearchResult(res);
            if (res.entries.length == 1 &&
                "true".equals(jEdit.getProperty("jindex.fastDisplay"))) {
                // one entry found for this searchstring, and the user
                // wishes fast display of help contents.
                showDocOnEntry(res.entries[0]);
            }
        }
    }


    private void setSearchResult(JIndex.SearchResult res) {
        keyselpos = res.keywordpos;
        if (keyselpos < 0) keyselpos = -keyselpos;
        keyList.setSelectedIndex(keyselpos);
        keyList.ensureIndexIsVisible(keyselpos);
        updateTopicList(res.entries);
    }


    private void updateTopicList(IndexEntry[] entries) {
        tlmodel.setEntries(entries);
        topicselpos = 0;
        topicList.setSelectedIndex(topicselpos);
        setStatusText(jEdit.getProperty("jindex.frame.status.countEntries",
            new Object[] {
                new Integer(entries == null ? 0 : entries.length),
                index.getKeywordAt(keyselpos)
            }
        ));
    }


    private void showDocOnSelectedEntry() {
        IndexEntry e = (IndexEntry) topicList.getSelectedValue();
        if (e != null)
            showDocOnEntry(e);
    }


    void showDocOnEntry(IndexEntry e) {
        setStatusText(jEdit.getProperty("jindex.frame.status.open",
            new Object[] { e.name } ));
        String url = e.getCompleteURL();
        Log.log(Log.DEBUG, this, "showDocOnEntry: " + url);
        openURL(url);
    }


    private void openURL(String url) {
        URL u;
        try {
            u = new URL(url);
        } catch(MalformedURLException e) {
            GUIUtilities.error(null, "jindex.error.docwrong",new Object[] { url, e });
            return;
        }
        openURL(u);
    }


    private void openURL(URL url) {
        // send a ViewURL message on the EditBus
        ViewURL vu = new ViewURL(this, null, url);
        EditBus.send(vu);

        if (!vu.isVetoed()) {
            // no EditBus component listened for this message.
            // Show an error:
            GUIUtilities.error(null, "jindex.error.noinfoviewer", null);
            return;
        }
    }


    private class TopicListModel extends AbstractListModel {

        private IndexEntry[] entries = null;


        public TopicListModel() {
            super();
        }


        public int getSize() { return entries == null ? 0 : entries.length; }
        public Object getElementAt(int i) {
            if (entries == null)
                return "";
            else
                return entries[i];
        }


        public void setEntries(IndexEntry[] entries) {
            this.entries = entries;
            fireContentsChanged(this, -1, -1);
        }

    } // inner class TopicListModel


    private class KeyListModel extends AbstractListModel {

        public int getSize() { return index == null ? 0 : index.getNumKeywords(); }
        public Object getElementAt(int i) { return index == null ? null : index.getKeywordAt(i); }
        public void fireContentsChanged() { fireContentsChanged(this, -1, -1); }

    } // inner class KeyListModel


    private class TopicListCellRenderer extends JLabel implements ListCellRenderer {

        public Component getListCellRendererComponent(
            JList list,              // the list component
            Object value,            // value to display
            int index,               // cell index
            boolean isSelected,      // is the cell selected
            boolean cellHasFocus)    // the list and the cell have the focus
        {
            IndexEntry e = (IndexEntry) value;
            this.setText(e.name);

            switch (e.type) {
                case IndexEntry.CLAZZ:           setIcon(classIcon); break;
                case IndexEntry.INTERFACE:       setIcon(interIcon); break;
                case IndexEntry.CONSTRUCTOR:     setIcon(constrIcon); break;
                case IndexEntry.METHOD: default: setIcon(methodIcon); break;
                case IndexEntry.FIELD:           setIcon(fieldIcon); break;
            }

            if (isSelected) {
                this.setBackground(list.getSelectionBackground());
                this.setForeground(list.getSelectionForeground());
            } else {
                this.setBackground(list.getBackground());
                this.setForeground(list.getForeground());
            }

            this.setFont(list.getFont());
             // Swing needs this, otherwise it doesn't paint
             // the background color:
            this.setOpaque(true);
            return this;
        }

    } // inner class TopicListCellRenderer

}
