/*
 * jEdit editor settings:
 * :mode=java:tabSize=4:indentSize=4:noTabs=true:maxLineLen=0:
 *
 * JIndexLoader.java
 * Copyright (C) 1999 Dirk Moebius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jindex;

import java.io.*;
import org.gjt.sp.jedit.*;
import org.gjt.sp.util.Log;
import com.microstar.xml.XmlException;


/**
 * this class is responsible for loading a JIndex index asynchronously and
 * holding it in a static member variable.
 */
public class JIndexLoader extends Thread implements EBComponent {

    public static final int STATUS_OK = 1;
    public static final int STATUS_NOT_EXISTS = 2;
    public static final int STATUS_LOAD_ERROR = 3;
    public static final int STATUS_INV_FILE = 4;
    public static final int STATUS_NO_SETTINGSDIR = 5;
    public static final int STATUS_NOT_LOADED = 6;
    public static final int STATUS_LOADING = 7;

    /** the filename of the serialized JIndex object */
    private static final String INDEXFILENAME = "jindex.ser";


    /** current JIndex instance. There can be only one index. */
    private static JIndex index = null;

    /**
     * the full qualified filename of the serialized JIndex object,
     * with directory etc.
     */
    private static String indexfilename;

    /** current status */
    private static int status;


    static {
        String homeDir = jEdit.getSettingsDirectory();
        if (homeDir == null) {
            status = STATUS_NO_SETTINGSDIR;
            GUIUtilities.error(null, "jindex.error.nohome", null);
        } else {
            String dirname = homeDir + File.separator + "jindex";
            File  indexdir = new File(dirname);
            indexfilename = dirname + File.separator + INDEXFILENAME;
            if (!indexdir.exists()) {
                indexdir.mkdirs();
                status = STATUS_NOT_EXISTS;
            } else {
                File f = new File(indexfilename);
                if (f.exists() && f.isFile() && f.canRead()) {
                    status = STATUS_NOT_LOADED;
                } else {
                    status = STATUS_NOT_EXISTS;
                }
            }
        }
    }


    /**
     * create a new JIndexLoader thread.
     */
    public JIndexLoader() {
        super("JIndexLoader");
        this.setPriority(NORM_PRIORITY - 1);
        EditBus.addToBus(this);
    }


    public void handleMessage(EBMessage message) {
        if (message instanceof JIndexChange) {
            if (message.getSource() != this) {
                // the index was changed by someone else, ie. by JIndexCreator
                index = ((JIndexChange)message).getJIndex();
                // assume the status is o.k.
                setStatus(STATUS_OK);
            }
        }
    }


    public void run() {
        if (status == STATUS_NOT_LOADED) {
            // inform listeners to display an info text.
            setStatus(STATUS_LOADING);
            Log.log(Log.DEBUG, this, "JIndexLoader thread start");

            // be nice to other processes and give them some time:
            yield();

            // load the JIndex
            try {
                FileReader fr = new FileReader(indexfilename);
                BufferedReader br = new BufferedReader(fr);
                JIndex newIndex = new JIndex();
                newIndex.readXML(br);
                br.close();
                setStatus(STATUS_OK);
                setIndex(newIndex);
            }
            catch (FileNotFoundException fnf) {
                setStatus(STATUS_NOT_EXISTS);
            }
            catch(XmlException xe) {
                setStatus(STATUS_INV_FILE);
                Log.log(Log.ERROR, this, xe.toString());
                GUIUtilities.error(null, "jindex.error.invalidfile",
                    new Object[] {
                        indexfilename,
                        xe.getMessage(),
                        new Integer(xe.getLine()),
                        new Integer(xe.getColumn())
                    });
                new ConfigureDialog(jEdit.getFirstView());
            }
            catch (Exception e) {
                setStatus(STATUS_LOAD_ERROR);
                Log.log(Log.ERROR, this, e.toString());
                GUIUtilities.error(null, "jindex.error.load",
                    new Object[] { indexfilename, e });
            }

            Log.log(Log.DEBUG, this, "JIndexLoader thread end; status=" + status);
        }
    }


    /**
     * Get the static JIndex instance. If the index is not yet loaded,
     * this will return null. (Use the EditBus StatusChange message to
     * know when the index is loaded completely.)
     */
    public JIndex getIndex() {
        if (status == STATUS_NOT_EXISTS) {
            // there is no index file: show informational message and
            // open JIndex creation dialog afterwards
            GUIUtilities.message(null, "jindex.createinfo", null);
            new ConfigureDialog(jEdit.getFirstView());
        }
        return index;
    }


    public static boolean indexExists() {
        return (status == STATUS_OK || status == STATUS_NOT_LOADED);
    }


    public static String getIndexFilename() {
        return indexfilename;
    }


    /** return the current status */
    public static int getStatus() {
        return status;
    }


    /**
     * sets the index and assumes, it is ok. Thus, the status will be set
     * to STATUS_OK. An JIndexChange message will be sent out on EditBus.
     * @param _index the new JIndex.
     */
    private synchronized void setIndex(JIndex _index) {
        index = _index;
        EditBus.send(new JIndexChange(this, index));
        setStatus(STATUS_OK);
    }


    /**
     * set a new status and send out a StatusChange message on EditBus.
     * @param _status the new status.
     */
    private synchronized void setStatus(int _status) {
        status = _status;
        EditBus.send(new StatusChange(this, status));
    }

}
