/*
 * jEdit editor settings:
 * :mode=java:tabSize=4:indentSize=4:noTabs=true:maxLineLen=0:
 *
 * MultipleEntriesDialog.java
 * Copyright (C) 2000 Dirk Moebius
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jindex;

import java.awt.*;
import java.awt.event.*;
import java.net.*;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;

import org.gjt.sp.jedit.*;
import org.gjt.sp.jedit.gui.EnhancedDialog;
import org.gjt.sp.jedit.msg.ShowTokenInfo;
import org.gjt.sp.jedit.msg.ViewURL;
import org.gjt.sp.util.Log;


/**
 * This is a chooser dialog, that is displayed, when there are multiple
 * entries for the current keyword. The user has to select one, or cancel
 * the dialog with ESC.
 */
public class MultipleEntriesDialog extends EnhancedDialog implements EBComponent {

    private static final Dimension TOPIC_LIST_SIZE = new Dimension(300,150);
    private static final Insets vertSpaceTop = new Insets(5,5,0,5);
    private static final Insets vertSpaceNone = new Insets(0,5,0,5);
    private static final Insets vertSpaceBoth = new Insets(5,5,5,5);
    private static final int fixedCellHeight = new JLabel("Mg_").getPreferredSize().height;

    private JIndex           index = null;
    private String           lastSearchString = null;
    private HelpfulJList     topicList;
    private TopicListModel   tlmodel;
    private JLabel           status;
    private int              topicselpos = -1;


    public MultipleEntriesDialog(View view, String searchString) {
        super(view,
              jEdit.getProperty("jindex.chooser.title",
                                new Object[] {searchString}),
              false);

        /************ create components ************/

        Font small = new Font("Dialog", Font.PLAIN, 10);

        // info label
        JLabel info = new JLabel(jEdit.getProperty("jindex.chooser.info"));
        info.setFont(small);

        // list of topics found
        tlmodel = new TopicListModel();
        topicList = new HelpfulJList(tlmodel);
        topicList.setFixedCellHeight(fixedCellHeight);
        topicList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        topicList.setCellRenderer(new TopicListCellRenderer());
        topicList.addMouseListener(new MouseAdapter() {
            public void mouseClicked(MouseEvent e) {
                if (e.getClickCount()==2) {
                    topicselpos = topicList.locationToIndex(e.getPoint());
                    ok();
                }
            }
        });

        // scrollpane for list of topics
        JScrollPane scrTopicList = new JScrollPane(topicList);
        scrTopicList.setPreferredSize(TOPIC_LIST_SIZE);

        // status label
        status = new JLabel(jEdit.getProperty("jindex.frame.status.label"));
        status.setFont(small);
        status.setBorder(new BevelBorder(BevelBorder.LOWERED));

        /************ general layout ************/

        getContentPane().setLayout(new BorderLayout());
        getContentPane().add(scrTopicList, BorderLayout.CENTER);
        getContentPane().add(info, BorderLayout.NORTH);
        getContentPane().add(status, BorderLayout.SOUTH);
        setSize(new Dimension(300,150));
        setLocationRelativeTo(view);
        GUIUtilities.loadGeometry(this, "jindex.chooser");

        /************ misc setup ************/

        // we like to be notified about JIndex and status changes:
        EditBus.addToBus(this);

        // load index
        JIndexLoader jil = new JIndexLoader();
        jil.start();
        index = jil.getIndex(); // note: the index may still be null

        // if index is not yet loaded, make window visible to see progress:
        if (index == null) {
            setVisible(true);
        }
    }


    // begin EBComponent implementation

    public void handleMessage(EBMessage message) {
        if (message instanceof StatusChange) {
            int newstatus = ((StatusChange)message).getStatus();
            if (newstatus == JIndexLoader.STATUS_LOADING) {
                setBusy(true);
                setStatusText(jEdit
                    .getProperty("jindex.frame.status.loading"));
            } else {
                setBusy(false);
                if (newstatus != JIndexLoader.STATUS_NOT_EXISTS &&
                    newstatus != JIndexLoader.STATUS_OK) {
                    // there was some error
                    setStatusText(jEdit.getProperty("jindex.frame.status.error"));
                }
            }
        }
        else if (message instanceof JIndexChange) {
            index = ((JIndexChange)message).getJIndex();
            Log.log(Log.DEBUG, this, "got index change");
            setStatusText(jEdit.getProperty("jindex.frame.status.ready"));
            tlmodel.setEntries(null);
            if (lastSearchString != null) search(lastSearchString);
        }
    }

    // end EBComponent implementation


    private void setStatusText(final String text) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                status.setText(text);
            }
        });
    }


    private void setBusy(boolean busy) {
        if (busy) {
            setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
        } else {
            setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        }
    }


    public void ok() {
        showDocOnSelectedEntry();
        setVisible(false);
    }


    public void cancel() {
        setVisible(false);
    }


    public void setVisible(boolean vis) {
        super.setVisible(vis);
        if (!vis)
            GUIUtilities.saveGeometry(this, "jindex.chooser");
    }


    public void search(String searchstring) {
        if (index == null) {
            setVisible(true);
            setStatusText(jEdit.getProperty("jindex.frame.status.notloaded"));
            lastSearchString = searchstring;
            return;
        }

        if (searchstring == null) {
            return; // nothing to search for
        }

        setTitle(jEdit.getProperty("jindex.chooser.title",
                                   new Object[] {searchstring}));
        // search in index for searchstring
        JIndex.SearchResult res = index.search(searchstring, true);
        if (res.entries == null) {
            // nothing found
            setStatusText(jEdit.getProperty(
                "jindex.frame.status.keywordnotfound",
                new Object[] { res.searchstring } ));
            Toolkit.getDefaultToolkit().beep();
        } else {
            setSearchResult(res.entries, index.getKeywordAt(res.keywordpos));
            if (res.entries.length == 1 &&
                "true".equals(jEdit.getProperty("jindex.fastDisplay"))) {
                showDocOnEntry(res.entries[0]);
            } else {
                setVisible(true);
                topicList.requestFocus();
            }
        }

        lastSearchString = null;
    }


    private void setSearchResult(IndexEntry[] entries, String keyword) {
        tlmodel.setEntries(entries);
        topicselpos = 0;
        topicList.setSelectedIndex(topicselpos);
        setStatusText(jEdit.getProperty("jindex.frame.status.countEntries",
            new Object[] { new Integer(entries == null ? 0 : entries.length),
                           keyword } )
        );
    }


    private void showDocOnSelectedEntry() {
        IndexEntry e = (IndexEntry) topicList.getSelectedValue();
        if (e != null)
            showDocOnEntry(e);
    }


    void showDocOnEntry(IndexEntry e) {
        setStatusText(jEdit.getProperty("jindex.frame.status.open",
                                        new Object[] { e.name } ));
        String url = e.getCompleteURL();
		Log.log(Log.DEBUG, this, "showDocOnEntry: " + url);
        openURL(url);
    }


    private void openURL(String url) {
        URL u;
        try {
            u = new URL(url);
        } catch(MalformedURLException e) {
            GUIUtilities.error(null, "jindex.error.docwrong",
                               new Object[] { url, e } );
            return;
        }
        openURL(u);
    }


    private void openURL(URL url) {
        // send a ViewURL message on the EditBus
        ViewURL vu = new ViewURL(this, null, url);
        EditBus.send(vu);

        if (!vu.isVetoed()) {
            // no EditBus component listened for this message.
            // Show an error:
            GUIUtilities.error(null, "jindex.error.noinfoviewer", null);
            return;
        }
    }


    // ======================================================================


    private class TopicListModel extends AbstractListModel {
        private IndexEntry[] entries = null;
        public TopicListModel() { }
        public int getSize() { return entries == null ? 0 : entries.length; }
        public Object getElementAt(int i) {
            if (entries == null)
                return "";
            else
                return entries[i];
        }
        public void setEntries(IndexEntry[] entries) {
            this.entries = entries;
            fireContentsChanged(this, -1, -1);
        }
    } // inner class TopicListModel


    // ======================================================================


    private class TopicListCellRenderer extends JLabel implements ListCellRenderer {
        public Component getListCellRendererComponent(
            JList list,              // the list component
            Object value,            // value to display
            int index,               // cell index
            boolean isSelected,      // is the cell selected
            boolean cellHasFocus)    // the list and the cell have the focus
        {
            IndexEntry e = (IndexEntry) value;
            this.setText(e.name);
            switch (e.type) {
                case IndexEntry.CLAZZ:       setIcon(JIndexHoldable.classIcon); break;
                case IndexEntry.INTERFACE:   setIcon(JIndexHoldable.interIcon); break;
                case IndexEntry.CONSTRUCTOR: setIcon(JIndexHoldable.constrIcon); break;
                default:
                case IndexEntry.METHOD:      setIcon(JIndexHoldable.methodIcon); break;
                case IndexEntry.FIELD:       setIcon(JIndexHoldable.fieldIcon); break;
            }
            if (isSelected) {
                this.setBackground(list.getSelectionBackground());
                this.setForeground(list.getSelectionForeground());
            } else {
                this.setBackground(list.getBackground());
                this.setForeground(list.getForeground());
            }
            this.setFont(list.getFont());
             // Swing needs this, otherwise it doesn't paint
             // the background color:
            this.setOpaque(true);
            return this;
        }
    } // inner class TopicListCellRenderer

}

